/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#pragma once

#include <boost/variant.hpp>
#include <QtPlugin>
#include "iradiostation.h"

template<typename>
class QFuture;

namespace Media
{
	struct AudioInfo;

	/** @brief Describes the data associated with a radio station restore.
	 *
	 * For now it can be just a list of AudioInfo structs.
	 *
	 * @sa AudioInfo
	 */
	using RadioRestoreVariant_t = boost::variant<QList<AudioInfo>>;

	/** @brief Describes the result of restoring a single radio station.
	 *
	 * The radio stations are identified by Media::RadioItemRole::RadioID.
	 */
	struct RadioRestoreResult
	{
		/** @brief The plugin this RadioRestoreResult corresponds to.
		 */
		QString PluginID_;

		/** @brief The radio station this RadioRestoreResult describes.
		 *
		 * This field is equal to some Media::RadioItemRole::RadioID.
		 */
		QString RadioID_;

		/** @brief The restore result itself.
		 */
		RadioRestoreVariant_t Restored_;
	};

	/** @brief A list of RadioRestoreResult structs.
	 */
	using RadiosRestoreResult_t = QList<RadioRestoreResult>;

	/** @brief Interface for radio station providers able to restore the
	 * radio stations between LeechCraft runs.
	 *
	 * An example of such a provider could be the VKontakte streamer. The
	 * links to the songs die after a certain amount of time, but still
	 * it's desirable to allow restoring the list of tracks.
	 *
	 * @sa IRadioStationProvider
	 */
	class IRestorableRadioStationProvider
	{
	protected:
		virtual ~IRestorableRadioStationProvider () {}
	public:
		/** @brief Initiates restoring the radio stations for the given
		 * IDs.
		 *
		 * The IDs are the ones exposed by the provider via the
		 * Media::RadioItemRole::RadioID model role.
		 *
		 * @param[in] ids The list of radio IDs to restore.
		 * @return The future providing a list of RadioRestoreResult
		 * structs.
		 *
		 * @sa RadioRestoreResult
		 */
		virtual QFuture<RadiosRestoreResult_t> RestoreRadioStations (const QStringList& ids) = 0;
	};
}

Q_DECLARE_INTERFACE (Media::IRestorableRadioStationProvider, "org.LeechCraft.Media.IRestorableRadioStationProvider/1.0")
