/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#pragma once

#include <QtPlugin>

namespace LeechCraft
{
namespace Azoth
{
	/** @brief Interface for entries that can modify messages bodies.
	 *
	 * This interface is to be implemented by entries (implementing
	 * ICLEntry) that can modify the bodies of the messages after the
	 * messages were received and handled by Azoth.
	 *
	 * When the body of the message is to be changed, the performJS()
	 * should be emitted with the corresponding JavaScript code that
	 * changes the message in the opened chat tabs. IMessage::GetBody()
	 * should also return the new message body after this signal.
	 *
	 * For example, a social network protocol plugin may emit an
	 * ICLEntry::gotMessage() signal when the message is received, but
	 * request additional information from the social network about the
	 * attachments, emitting IUpdatableChatEntry::performJS() after that
	 * info is fetched.
	 *
	 * @sa ICLEntry
	 */
	class IUpdatableChatEntry
	{
	public:
		virtual ~IUpdatableChatEntry () {}
	protected:
		/** @brief Emitted when a message body is to be changed.
		 *
		 * This signal should be emitted by entries when the contents of
		 * one or more messages should be changed. The corresponding
		 * \em js code is executed in an opened chat tab corresponding to
		 * the entry (if any).
		 *
		 * Typically, a message will contain a <code>div</code> or a
		 * <code>span</code> element with an unique  <code>id</code>
		 * attribute which is to be manipulated by the \em js code.
		 *
		 * @note This function is expected to be a signal.
		 *
		 * @param[out] js The code to execute in a chat tab corresponding
		 * to this entry, if any.
		 */
		virtual void performJS (const QString& js) = 0;
	};
}
}
Q_DECLARE_INTERFACE (LeechCraft::Azoth::IUpdatableChatEntry,
		"org.Deviant.LeechCraft.Azoth.IUpdatableChatEntry/1.0")
