/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "albumartfetcher.h"
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QNetworkAccessManager>
#include <QDomDocument>
#include <QStringList>
#include <QFuture>
#include <util/network/handlenetworkreply.h>
#include <util/sll/visitor.h>
#include <util/threads/futures.h>
#include "util.h"

namespace LeechCraft
{
namespace Lastfmscrobble
{
	AlbumArtFetcher::AlbumArtFetcher (const Media::AlbumInfo& albumInfo, ICoreProxy_ptr proxy, QObject *parent)
	: QObject (parent)
	{
		Promise_.reportStarted ();

		const QMap<QString, QString> params
		{
			{ "artist", albumInfo.Artist_ },
			{ "album", albumInfo.Album_ },
			{ "autocorrect", "1" }
		};

		const auto reply = Request ("album.getInfo", proxy->GetNetworkAccessManager (), params);
		Util::HandleReplySeq<Util::ErrorInfo<QString>> (reply, this) >>
				Util::Visitor
				{
					[this] (const QString& error) { Util::ReportFutureResult (Promise_, error); },
					[this] (const QByteArray& result) { HandleReplyFinished (result); }
				}.Finally ([this] { deleteLater (); });
	}

	QFuture<Media::IAlbumArtProvider::Result_t> AlbumArtFetcher::GetFuture ()
	{
		return Promise_.future ();
	}

	void AlbumArtFetcher::HandleReplyFinished (const QByteArray& data)
	{
		QDomDocument doc;
		if (!doc.setContent (data))
		{
			Util::ReportFutureResult (Promise_, "Unable to parse reply.");
			return;
		}

		const auto& elems = doc.elementsByTagName ("image");

		static const QStringList sizes
		{
			"mega",
			"extralarge",
			"large",
			"medium",
			"small",
			""
		};

		for (const auto& size : sizes)
			for (int i = 0; i < elems.size (); ++i)
			{
				const auto& elem = elems.at (i).toElement ();
				const auto& text = elem.text ();
				if (elem.attribute ("size") == size && !text.isEmpty ())
				{
					Util::ReportFutureResult (Promise_, QList { QUrl { elem.text () } });
					break;
				}
			}
	}
}
}
