/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "tagsfetchmanager.h"
#include <QFuture>
#include <util/threads/futures.h>
#include <util/sll/delayedexecutor.h>
#include <interfaces/lmp/mediainfo.h>
#include <interfaces/media/itagsfetcher.h>
#include "filesmodel.h"

namespace LeechCraft
{
namespace LMP
{
namespace Graffiti
{
	TagsFetchManager::TagsFetchManager (const QStringList& paths,
			Media::ITagsFetcher *prov, FilesModel *filesModel, QObject *parent)
	: QObject (parent)
	, FilesModel_ (filesModel)
	, FetchedTags_ (0)
	, TotalTags_ (paths.size ())
	{
		for (const auto& path : paths)
			Util::Sequence (this, prov->FetchTags (path)) >>
					[this, path] (const Media::AudioInfo& result) { handleTagsFetched (path, result); };

		Util::ExecuteLater ([this] { emit tagsFetchProgress (0, TotalTags_, this); });
	}

	namespace
	{
		template<typename T>
		bool IsEmptyData (const T& val)
		{
			if constexpr (std::is_default_constructible<T> {})
				return val == T {};
			else
				static_assert (std::is_same<T, struct Error> {}, "not a default-constructible data type");
		}

		template<typename F>
		void UpgradeInfo (MediaInfo& info, MediaInfo& other, F getter)
		{
			static_assert (std::is_lvalue_reference<std::result_of_t<F (MediaInfo&)>> {},
					"functor doesn't return an lvalue reference");

			auto& data = std::invoke (getter, info);
			const auto& otherData = std::invoke (getter, other);
			if (!IsEmptyData (otherData) && IsEmptyData (data))
				data = otherData;
		}
	}

	void TagsFetchManager::handleTagsFetched (const QString& filename, const Media::AudioInfo& result)
	{
		emit tagsFetchProgress (++FetchedTags_, TotalTags_, this);

		const auto& index = FilesModel_->FindIndex (filename);
		if (!index.isValid ())
			return;

		auto newInfo = MediaInfo::FromAudioInfo (result);

		auto info = index.data (FilesModel::Roles::MediaInfoRole).value<MediaInfo> ();
		UpgradeInfo (info, newInfo, &MediaInfo::Title_);
		UpgradeInfo (info, newInfo, &MediaInfo::Artist_);
		UpgradeInfo (info, newInfo, &MediaInfo::Album_);
		UpgradeInfo (info, newInfo, &MediaInfo::Year_);
		UpgradeInfo (info, newInfo, &MediaInfo::TrackNumber_);
		UpgradeInfo (info, newInfo, &MediaInfo::Genres_);
		FilesModel_->UpdateInfo (index, info);

		emit tagsFetched (filename);

		if (FetchedTags_ == TotalTags_)
			emit finished (true);
	}
}
}
}
