/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2010-2012  Oleg Linkin
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "syncer.h"
#include <QtDebug>
#include <QFileInfo>
#include <QDir>
#include <QCryptographicHash>
#include "xmlsettingsmanager.h"
#include "drivemanager.h"

namespace LeechCraft
{
namespace NetStoreManager
{
namespace GoogleDrive
{
	Syncer::Syncer (DriveManager *dm, QObject *parent)
	: QObject (parent)
	, LastChangesId_ (XmlSettingsManager::Instance ()
			.Property ("LastChangesId", 0).toLongLong ())
	, DM_ (dm)
	{
	}

	void Syncer::CheckRemoteStorage ()
	{
		connect (DM_,
				SIGNAL (gotChanges (QList<DriveChanges>, qlonglong)),
				this,
				SLOT (handleGotDriveChanges (QList<DriveChanges>, qlonglong)));
		DM_->RequestFileChanges (LastChangesId_);
	}

	void Syncer::CheckLocalStorage (const QStringList& paths,
			const QString& baseDir)
	{
		BaseDir_ = baseDir;
		Paths_ =  paths;
		std::sort (Paths_.begin (), Paths_.end ());
		connect (DM_,
				SIGNAL (gotFiles (QList<DriveItem>)),
				this,
				SLOT (handleGotFiles (QList<DriveItem>)));
		DM_->RefreshListing ();
	}

	void Syncer::ContinueLocalStorageChecking ()
	{
		if (Paths_.isEmpty ())
		{
			//TODO checking finished
			qDebug () << "finished";
			return;
		}

		QString path = Paths_.takeAt (0);
		QFileInfo info (path);
		DriveItem parentItem = RealPath2Item_.value (info.dir ().absolutePath ());

		bool found = false;
		DriveItem currentItem;
		for (const auto& item : Items_)
		{
			if (item.IsFolder_ != info.isDir () ||
					item.Name_ != info.fileName () ||
					item.Labels_ & DriveItem::ILRemoved ||
					item.ParentId_ != parentItem.Id_)
				continue;

			currentItem = item;
			found = true;
			Items_.removeOne (item);
			break;
		}

		if (found)
		{
			if (QFileInfo (path).exists ())
			{
				RealPath2Item_ [path] = currentItem;
				if (!info.isDir ())
				{
					QFile file (path);
					if (file.open (QIODevice::ReadOnly))
					{
						if (currentItem.Md5_ != QCryptographicHash::hash (file.readAll (), QCryptographicHash::Md5).toHex ())
						{
							//TODO update
							qDebug () << currentItem.Id_ << currentItem.Name_ << "need update";
						}
					}
				}
			}
			else
				DM_->MoveEntryToTrash (currentItem.Id_);
			ContinueLocalStorageChecking ();
		}
		else
		{
			connect (DM_,
					SIGNAL (gotNewItem (DriveItem)),
					this,
					SLOT (handleGotNewItem (DriveItem)));
			RealPathQueue_ << path;
			!info.isDir () ?
				DM_->Upload (path, QStringList () << parentItem.Id_) :
				DM_->CreateDirectory (info.fileName (), parentItem.Id_);
		}
	}

	void Syncer::handleGotDriveChanges (const QList<DriveChanges>& changes, qlonglong id)
	{
		disconnect (DM_,
			SIGNAL (gotChanges (QList<DriveChanges>, qlonglong)),
			this,
			SLOT (handleGotDriveChanges (QList<DriveChanges>, qlonglong)));

		XmlSettingsManager::Instance ().setProperty ("LastChangesId", id);
		LastChangesId_ = id;
	}

	void Syncer::handleGotFiles (const QList<DriveItem>& files)
	{
		Items_ = files;
		disconnect (DM_,
				SIGNAL (gotFiles (QList<DriveItem>)),
				this,
				SLOT (handleGotFiles (QList<DriveItem>)));

		const QString& rootName = QFileInfo (BaseDir_).fileName ();
		DriveItem rootItem;
		bool found = false;
		for (const auto& item : files)
		{
			if (!item.ParentIsRoot_ ||
					!item.IsFolder_ ||
					item.Name_ != rootName ||
					item.Labels_ & DriveItem::ILRemoved)
				continue;

			rootItem = item;
			found = true;
			Items_.removeOne (rootItem);
			break;
		}

		if (!found)
		{
			connect (DM_,
					SIGNAL (gotNewItem (DriveItem)),
					this,
					SLOT (handleGotNewItem (DriveItem)));
			RealPathQueue_ << BaseDir_;
			DM_->CreateDirectory (rootName);
		}
		else
		{
			if (QFileInfo (BaseDir_).exists ())
				RealPath2Item_ [BaseDir_] = rootItem;
			else
				DM_->MoveEntryToTrash (rootItem.Id_);

			ContinueLocalStorageChecking ();
		}
	}

	void Syncer::handleGotNewItem (const DriveItem& item)
	{
		disconnect (DM_,
				SIGNAL (gotNewItem (DriveItem)),
				this,
				SLOT (handleGotNewItem (DriveItem)));
		if (!RealPathQueue_.isEmpty ())
			RealPath2Item_ [RealPathQueue_.dequeue ()] = item;

		Items_ << item;
		qDebug () << "created entry: " << item.Name_ << RealPath2Item_.key (item);
		ContinueLocalStorageChecking ();
	}
}
}
}

