/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "viewsmanager.h"
#include <QAction>
#include <QtDebug>
#include "effectprocessor.h"
#include "xmlsettingsmanager.h"
#include "scripthandler.h"

namespace LeechCraft
{
namespace Poshuku
{
namespace DCAC
{
	ViewsManager::ViewsManager (IPluginsManager *ipm, QObject *parent)
	: QObject { parent }
	, ScriptHandler_ { new ScriptHandler { ipm, this } }
	{
		XmlSettingsManager::Instance ().RegisterObject ({
					"NightModeThreshold",
					"ReduceLightnessFactor",
					"SingleEffect",
					"ColorTemperature"
				},
				this, "handleEffectsChanged");

		connect (ScriptHandler_,
				SIGNAL (effectsListChanged ()),
				this,
				SLOT (handleEffectsChanged ()));
	}

	void ViewsManager::AddView (QWidget *view)
	{
		const auto effect = new EffectProcessor { view };
		view->setGraphicsEffect (effect);

		View2Effect_ [view] = effect;

		connect (view,
				SIGNAL (destroyed (QObject*)),
				this,
				SLOT (handleViewDestroyed (QObject*)));

		const auto enable = XmlSettingsManager::Instance ()
				.property ("EnableNightModeByDefault").toBool ();
		effect->setEnabled (enable);

		const auto enableAct = new QAction { tr ("Night mode"), view };
		view->addAction (enableAct);
		enableAct->setShortcut (QString { "Ctrl+Shift+I" });
		enableAct->setCheckable (true);
		enableAct->setChecked (enable);
		connect (enableAct,
				SIGNAL (toggled (bool)),
				effect,
				SLOT (setEnabled (bool)));
		View2EnableAction_ [view] = enableAct;

		effect->SetEffects (GetCurrentEffects ());
	}

	QAction* ViewsManager::GetEnableAction (QWidget *view) const
	{
		return View2EnableAction_.value (view);
	}

	QList<Effect_t> ViewsManager::GetCurrentEffects ()
	{
		const auto effectStr = XmlSettingsManager::Instance ()
				.property ("SingleEffect").toString ();
		if (effectStr == "Invert")
		{
			const auto threshold = XmlSettingsManager::Instance ()
					.property ("NightModeThreshold").toInt ();
			return { InvertEffect { threshold } };
		}
		else if (effectStr == "ReduceLightness")
		{
			const auto factor = XmlSettingsManager::Instance ()
					.property ("ReduceLightnessFactor").toDouble ();
			return { LightnessEffect { factor } };
		}
		else if (effectStr == "ColorTemp")
		{
			const auto temp = XmlSettingsManager::Instance ()
					.property ("ColorTemperature").toInt ();
			return { ColorTempEffect { temp } };
		}
		else if (effectStr == "Script")
		{
			const auto& scriptPath = XmlSettingsManager::Instance ()
					.property ("ScriptPath").toString ();
			ScriptHandler_->SetScriptPath (scriptPath);
			return ScriptHandler_->GetEffects ();
		}
		else
		{
			qWarning () << Q_FUNC_INFO
					<< "unknown effect"
					<< effectStr;
			return {};
		}
	}

	void ViewsManager::handleViewDestroyed (QObject *view)
	{
		View2Effect_.remove (view);
		View2EnableAction_.remove (view);
	}

	void ViewsManager::handleEffectsChanged ()
	{
		const auto& effects = GetCurrentEffects ();
		for (const auto proc : View2Effect_)
			proc->SetEffects (effects);
	}
}
}
}
