/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "requestinterceptor.h"
#include <QWebEngineUrlRequestInfo>
#include <QtDebug>
#include <util/sll/visitor.h>
#include <util/sll/slotclosure.h>
#include "customwebview.h"

namespace LeechCraft::Poshuku::WebEngineView
{
	namespace
	{
		IInterceptableRequests::ResourceType ConvertResourceType (QWebEngineUrlRequestInfo::ResourceType type)
		{
#define HANDLE(x)	case QWebEngineUrlRequestInfo::ResourceType##x: \
						return IInterceptableRequests::ResourceType::x;

			switch (type)
			{
			HANDLE (MainFrame)
			HANDLE (SubFrame)
			HANDLE (Stylesheet)
			HANDLE (Script)
			HANDLE (Image)
			HANDLE (FontResource)
			HANDLE (SubResource)
			HANDLE (Object)
			HANDLE (Media)
			HANDLE (Worker)
			HANDLE (SharedWorker)
			HANDLE (Prefetch)
			HANDLE (Favicon)
			HANDLE (Xhr)
			HANDLE (Ping)
			HANDLE (ServiceWorker)
			HANDLE (CspReport)
			HANDLE (PluginResource)
			HANDLE (Unknown)
			case QWebEngineUrlRequestInfo::ResourceTypeLast:
				qWarning () << Q_FUNC_INFO << "got ResourceTypeLast";
				return IInterceptableRequests::ResourceType::Unknown;
			}
#undef HANDLE

			return IInterceptableRequests::ResourceType::Unknown;
		}
	}

	void RequestInterceptor::interceptRequest (QWebEngineUrlRequestInfo& info)
	{
		const auto& pageUrl = info.firstPartyUrl ();
		const auto page = [&] () -> boost::optional<IWebView*>
		{
			for (const auto view : Views_)
				if (pageUrl == view->GetUrl ())
					return view;
			return {};
		} ();
		IInterceptableRequests::RequestInfo convertedInfo
		{
			info.requestUrl (),
			pageUrl,
			IInterceptableRequests::NavigationType::Unknown,
			ConvertResourceType (info.resourceType ()),
			page
		};

		for (const auto& interceptor : Interceptors_)
		{
			const auto shouldBlock = Util::Visit (interceptor (convertedInfo),
					[] (IInterceptableRequests::Allow) { return false; },
					[&] (const IInterceptableRequests::Redirect& r)
					{
						info.redirect (r.NewUrl_);
						return false;
					},
					[] (IInterceptableRequests::Block) { return true; });

			if (shouldBlock)
				info.block (true);
		}
	}

	void RequestInterceptor::Add (const IInterceptableRequests::Interceptor_t& interceptor)
	{
		Interceptors_ << interceptor;
	}

	void RequestInterceptor::RegisterView (CustomWebView *view)
	{
		Views_ << view;

		new Util::SlotClosure<Util::DeleteLaterPolicy>
		{
			[this, view] { Views_.removeOne (view); },
			view,
			SIGNAL (destroyed ()),
			this
		};
	}
}
