/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "preludetest.h"
#include <QtTest>
#include <prelude.h>
#include <qtutil.h>

QTEST_MAIN (LeechCraft::Util::PreludeTest)

namespace LeechCraft
{
namespace Util
{
	namespace
	{
		QMap<int, QString> GetSimpleMap ()
		{
			return { { 0, "aaa" }, { 1, "bbb" }, { 2, "ccc" }};
		}
	}

	void PreludeTest::testMapList ()
	{
		QList<int> list { 1, 2, 3 };
		const auto& otherList = Map (list, [] (int v) { return QString::number (v); });

		QCOMPARE (otherList, (QStringList { "1", "2", "3" }));
	}

	void PreludeTest::testMapMap ()
	{
		const auto& map = GetSimpleMap ();
		const auto& otherList = Map (map, [] (const QString& v) { return v.size (); });

		QCOMPARE (otherList, (QList<int> { 3, 3, 3 }));
	}

	void PreludeTest::testMapStringList ()
	{
		const QStringList list { "aaa", "bbb", "ccc" };
		const auto& result = Map (list, [] (const QString& s) { return s.size (); });

		QCOMPARE (result, (QList<int> { 3, 3, 3 }));
	}

	void PreludeTest::testMapMapStlized ()
	{
		const auto& map = GetSimpleMap ();
		const auto& list = Map (Stlize (map), [] (const std::pair<int, QString>& pair) { return pair.second; });

		QCOMPARE (list, QStringList { map.values () });
	}

	void PreludeTest::testMapMember ()
	{
		struct Test
		{
			int m_a;
			int m_b;
		};

		const QList<Test> tests { { 1, 2 }, { 2, 4 }, { 3, 6 } };
		const auto& ints = Map (tests, &Test::m_a);

		QCOMPARE (ints, (QList<int> { 1, 2, 3 }));
	}

	void PreludeTest::testMapMemberFunction ()
	{
		struct Test
		{
			int m_a;

			int GetA () const
			{
				return m_a;
			}
		};

		const QList<Test> tests { { 1 }, { 2 }, { 3 } };
		const auto& ints = Map (tests, &Test::GetA);

		QCOMPARE (ints, (QList<int> { 1, 2, 3 }));
	}

	void PreludeTest::testConcatLists ()
	{
		QList<QList<int>> listOfLists
		{
			{ 1, 2 },
			{ 3 },
			{ 4, 5, 6 }
		};

		const auto& concat = Concat (listOfLists);
		QCOMPARE (concat, (QList<int> { 1, 2, 3, 4, 5, 6 }));
	}

	void PreludeTest::testConcatSets ()
	{
		QList<QSet<int>> listOfSets
		{
			{ 1, 2, 3 },
			{ 3, 4 },
			{ 4, 5, 6 }
		};

		const auto& concat = Concat (listOfSets);
		QCOMPARE (concat, (QSet<int> { 1, 2, 3, 4, 5, 6 }));
	}
}
}
